 #include "algo.h"
       
extern double THRESHOLD;
extern struct val_index * vpart;
extern int X,Y;
//BRANCH_DIST
void Branch_Dist(Graph g, int i, int i1, int j, double dist, double *finaldist)
{
	int l,sito;
	double d = 0;
	

	for( l = 0; l<g.Node[i].nn; ++l ){
		if ( g.Edge[g.Node[i].neigh[l]].s1 == i )
			sito = g.Edge[g.Node[i].neigh[l]].s2;
		else
			sito = g.Edge[g.Node[i].neigh[l]].s1;
		if ( sito != i1 ){			
			if ( sito == j ){
				*finaldist = g.Edge[g.Node[i].neigh[l]].dist + dist;
				return;
			}
			else {
				d  = dist+g.Edge[g.Node[i].neigh[l]].dist;
				Branch_Dist(g,sito,i,j,d , finaldist);		
			}
		}
	}

	return;
}
/****************************************************************/


//BRANCH_DIST_TOPOLOGICAL
void Branch_Dist_Topological(Graph g, int i, int i1, int j, int dist, int *finaldist)
{
	int l,sito;
	double d = 0;
	

	for( l = 0; l<g.Node[i].nn; ++l ){
		if ( g.Edge[g.Node[i].neigh[l]].s1 == i )
			sito = g.Edge[g.Node[i].neigh[l]].s2;
		else
			sito = g.Edge[g.Node[i].neigh[l]].s1;
		if ( sito != i1 ){			

			if ( sito == j ){
				*finaldist = 1 + dist;
				return;

			}
			else {
				d  = dist + 1;
				Branch_Dist_Topological(g,sito,i,j,d , finaldist);		
			}
		}
	}

	return;
}
/*************************************************************/


//COMPUTE_LEAVES_DIST
double ** Compute_Leaves_Dist ( Graph g )
{

	int i,j;
	double ** d;
	double a;

	d = (double ** ) malloc ( g.nleaves * sizeof( double * ) );


	for( i = 0; i< g.nleaves; ++i ){
		d[i] = (double *) malloc ( g.nleaves * sizeof( double ));
		d[i][i] = 0;
	}


	for( i = 0; i <g.nleaves-1; ++i ){
		for( j = i+1; j<g.nleaves; ++j ){
			a = 0;
			Branch_Dist(g,i,i,j,0.0,&a);
			d[i][j] = a;
			d[j][i] = d[i][j];
			
		}
	}
	return d;


}
/**********************************************************/


//COMPUTE_LEAVES_DIST_TOPOLOGICAL
void Compute_Leaves_Dist_Topological ( Graph g , int ** d)
{

	int i,j;
	int a;

	for( i = 0; i< g.nleaves; ++i ){
		d[i][i] = 0;
	}

	for( i = 0; i <g.nleaves-1; ++i ){
		for( j = i+1; j<g.nleaves; ++j ){
			a = 0;
	
			Branch_Dist_Topological(g,i,i,j,0.0,&a);

			d[i][j] = a;
			d[j][i] = d[i][j];
			
		}
	}

}
/********************************************************/


//CALCOLALENGHTPAULPIN
double CalcolaLengthPauplin(Graph g, int **d){

	int i,j;
	double length;

	length=0;
	for( i = 0; i< g.nleaves; ++i ){
		for( j = 0; j< g.nleaves; ++j ){
			length+=g.leavesdist[i][j]*1./pow(2.0,d[i][j]);
		}
	}
	return length;

}
/********************************************************/


//RETURNLINKNEIGHBORS
void ReturnLinkNeighbors( Graph g, int edge, int *s1, int *s2, int *s11, int *s12,  int * s21, int *s22, int *e11, int *e12, int *e21, int *e22) 
{
	int i,counter;	
	int v[2];
	int e[2];


	*s1 = g.Edge[edge].s1;
	*s2 = g.Edge[edge].s2;

	counter = 0;

	for( i = 0; i< g.Node[*s1].nn; ++i )
		if ( g.Node[*s1].neigh[i] != edge ){
			e[counter] =  g.Node[*s1].neigh[i];
			if ( g.Edge[g.Node[*s1].neigh[i]].s1 != *s1 )
				v[counter] = g.Edge[g.Node[*s1].neigh[i]].s1;
				
			else 
				v[counter] = g.Edge[g.Node[*s1].neigh[i]].s2;
			counter++;
		}
	assert(counter==2);
	*s11 = v[0];
	*s12 = v[1];
	*e11 = e[0];
	*e12 = e[1];

	counter = 0;
	for( i = 0; i< g.Node[*s2].nn; ++i )
		if ( g.Node[*s2].neigh[i] != edge ){
			e[counter] =  g.Node[*s2].neigh[i];
			if ( g.Edge[g.Node[*s2].neigh[i]].s1 != *s2 )
				v[counter] = g.Edge[g.Node[*s2].neigh[i]].s1;
			else 
				v[counter] = g.Edge[g.Node[*s2].neigh[i]].s2;
			counter++;
		}


	assert(counter==2);
	*s21 = v[0];
	*s22 = v[1];
	*e21 = e[0];
	*e22 = e[1];
	
} 
/******************************************************************/


//PRINTGRAPHDOT
void PrintGraphDot( Graph g )
{
	int i;
	FILE * fp;

	fp = fopen("plotgraph.dot","w");
	fprintf(fp, "digraph{");
	fprintf(fp,"size=\"15,10\";\n");

	for( i=0; i<g.nnodes; ++i )
		fprintf(fp,"%d [label=\"%i\"];\n",i,i);

	for( i=0; i<g.nedges; ++i )
		fprintf(fp,"%i -> %i  [label=e%d];\n", g.Edge[i].s1, g.Edge[i].s2,i);
	
	fprintf(fp, "};\n");
	fclose(fp);
}
/****************************************************************/


//COMPUTEPAUPLIN
void ComputePauplin(Graph g , int i, int e, double dist, double * finaldist, double * vecfinaldist, int *vecleaves, int * nleaves)
{

	int l,sito;
	double d = 1;





	if (g.Node[i].nn == 1 ){
		*finaldist = dist;
		vecfinaldist[*nleaves] = *finaldist;
		vecleaves[*nleaves] = i ;
		(*nleaves)++;
		return;
	}

	else{	
		for( l = 0; l<g.Node[i].nn; ++l ){
			if (g.Node[i].neigh[l] != e){
				if ( g.Edge[g.Node[i].neigh[l]].s1 == i )
					sito = g.Edge[g.Node[i].neigh[l]].s2;
				else
					sito = g.Edge[g.Node[i].neigh[l]].s1;
				if ( sito != i ){
					if ( g.Node[sito].nn == 1 ){
						*finaldist = 2*dist;
						vecfinaldist[*nleaves] = *finaldist;
						vecleaves[*nleaves] = sito;
						(*nleaves)++;
						
					}
					else{
						d  = dist *2;
						ComputePauplin( g, sito, g.Node[i].neigh[l] , d, finaldist, vecfinaldist, vecleaves, nleaves );
					}
				}
			}
			
		}

		return;
	}
	return;
}
/**********************************************************/



//COMPUTELEAVESFROMSITE
void ComputeLeavesFromSite(Graph g , int i, int * vecleaves, int * nleaves)
{

	int sito,j;


	if (g.Node[i].nn == 1 ){		
		vecleaves[*nleaves]=i ;
		(*nleaves)++;
		return;
	}

	else{	
		
				for(j=0;j<2;j++){
					sito = g.Node[i].sons[j];
			                if(sito>=0){	
					if ( g.Node[sito].nn == 1 ){
						vecleaves[*nleaves]=sito;
						(*nleaves)++;
					
					}
					else{
						
						ComputeLeavesFromSite( g, sito, vecleaves, nleaves );}
					}}
		
	  }
	
return;
}
/*****************************************************/



//CSTRING_CMP
int cstring_cmp(const void *a, const void *b)
{
    const char **ia = (const char **)a;
    const char **ib = (const char **)b;
    return strcmp(*ia, *ib);
	/* strcmp functions works exactly as expected from
	comparison function */
}
/****************************************************/




//COMP
int comp(void const * a, void const * b) {
        struct val_index const *aa = a;
        struct val_index const *bb = b;
        if (aa->val > bb->val)
                return 1;
        if (aa->val < bb->val)
                return -1;
        return 0;
}
/***************************************************/




//COPYVAL
void CopyVal(char **orig, struct val_index * vi, int dim){

	int i;
	for(i=0;i<dim;i++){
		strcpy(vi[i].val,orig[i]);
		vi[i].index=i;
	}

}
/***************************************************/




//COPYVALINVERSE
void CopyValInverse(char **dest, struct val_index * vi, int dim){

	int i;
	for(i=0;i<dim;i++){
		strcpy(dest[i],vi[i].val);
	}

}
/***************************************************/


//COPYVALINT
void CopyValInt(int *orig, struct int_index * vi, int dim){

        int i;
        for(i=0;i<dim;i++){
                vi[i].val=orig[i];
                vi[i].index=i;
        }

}
/***************************************************/



//COPYVALINVERSEINT
void CopyValInverseInt(int *dest, struct int_index * vi, int dim){

        int i;
        for(i=0;i<dim;i++){
                dest[i]=vi[i].val;
        }

}
/***************************************************/




//NEWICKRECURSION
void NewickRecursion( Graph g, int site, int node, int first )
{
	int i,countflag,vici;
	int nodehelp[2],link1,link2;
	
	countflag=0; vici=0;
	for(i=0; i < g.Node[node].nneigh; i++){
		if(g.Node[node].siteneigh[i] != site){
			nodehelp[vici]=g.Node[node].siteneigh[i];
			countflag+=g.Node[nodehelp[vici]].newick;
			if(!g.Node[nodehelp[vici]].newick && nodehelp[vici]!=first )
				NewickRecursion( g, node,  nodehelp[vici],first);
			vici++;
		}
	}
	//printf("%d %d %d %d nodo: %d\n",nodehelp[0],nodehelp[1],g.Node[nodehelp[0]].siteneigh[0],g.Node[nodehelp[1]].siteneigh[0],node);
	    for( i = 0; i<g.nedges; ++i ){
		if(g.Edge[i].s1==node && g.Edge[i].s2==nodehelp[0]){link1=i;}
                if(g.Edge[i].s2==node && g.Edge[i].s1==nodehelp[0]){link1=i;}
		if(g.Edge[i].s1==node && g.Edge[i].s2==nodehelp[1]){link2=i;}
                if(g.Edge[i].s2==node && g.Edge[i].s1==nodehelp[1]){link2=i;}
            
	}

      if(countflag == 2){
		sprintf(g.Node[node].labelnewick,"(%s:%.3lf,%s:%.3lf)",g.Node[nodehelp[0]].labelnewick,g.Edge[link1].dist,g.Node[nodehelp[1]].labelnewick,g.Edge[link2].dist);

		g.Node[node].newick = 1;
	
	}
	
}
/***************************************************/




//PRINTNEWICKFORMAT
void PrintNewickFormat( Graph g , int **disttopo ,FILE *fpgrafo)
{
	
	int i,x=-1,y=-1;
        int id1,id2,link1;
	int site1,site2=-1,count;
        

x=X;
y=Y;

	for( i = 0; i<g.nnodes; ++i ){
		g.Node[i].nneigh=0;
	}

	for( i = 0; i<g.nedges; ++i ){
		id1 = g.Edge[i].s1;
		id2 = g.Edge[i].s2;
		g.Node[id1].siteneigh[g.Node[id1].nneigh]=id2;
		g.Node[id1].nneigh++;
		g.Node[id2].siteneigh[g.Node[id2].nneigh]=id1;
		g.Node[id2].nneigh++;
	}

	for( i = 0; i<g.nnodes; ++i ){
		assert(g.Node[i].nneigh <= 3);

		
	}


	assert(g.Node[x].siteneigh[0]==g.Node[y].siteneigh[0]);
	site1= g.Node[x].siteneigh[0];

     /* for( i = 0; i<g.nedges; ++i ){
		if(g.Edge[i].s1==x || g.Edge[i].s2==x){link1=i;}
                if(g.Edge[i].s1==y || g.Edge[i].s2==y){link2=i;}
	}
        printf("%d %d: %d %d\n",link1,link2,g.Node[x].neigh[0],g.Node[y].neigh[0]); */

	sprintf(g.Node[site1].labelnewick,"(%s:%.3lf,%s:%.3lf)",g.Node[x].labelnewick,g.Edge[g.Node[x].neigh[0]].dist,g.Node[y].labelnewick,g.Edge[g.Node[y].neigh[0]].dist);

	for(i=0;i<g.Node[site1].nn ;i++)
		if(g.Node[site1].siteneigh[i] != x && g.Node[site1].siteneigh[i] != y)
			site2= g.Node[site1].siteneigh[i];
	count=g.nleaves;
	while(count<g.nnodes-1){
		NewickRecursion( g, site1, site2 ,site1);
		count=g.nleaves;
		for( i = g.nleaves; i <g.nnodes; ++i ){
			count+=g.Node[i].newick;
		}
	}
    for( i = 0; i<g.nedges; ++i ){
		if(g.Edge[i].s1==site1 && g.Edge[i].s2==site2){link1=i;}
                if(g.Edge[i].s1==site2 && g.Edge[i].s2==site1){link1=i;}
            
	}
	fprintf(fpgrafo,"(%s:%.3lf,%s:%.3lf,%s:%.3lf);\n",g.Node[x].labelnewick,g.Edge[g.Node[x].neigh[0]].dist,g.Node[y].labelnewick,g.Edge[g.Node[y].neigh[0]].dist,g.Node[site2].labelnewick,g.Edge[link1].dist);

	}
/***************************************************/



//BUILD_FAMILY
void build_family(Graph g , int i, int e, int *vecleaves, int * nleaves)
{

	int l,sito,nson;
	if (g.Node[i].nn == 1 ){		
		vecleaves[*nleaves] = i ;
	    (*nleaves)++;
		
		if ( g.Edge[g.Node[i].neigh[0]].s1 == i )
			sito = g.Edge[g.Node[i].neigh[0]].s2;
		else
			sito = g.Edge[g.Node[i].neigh[0]].s1;
		g.Node[i].father=sito;
		return;
	}
	
	else{	
		nson=0;
		for( l = 0; l<g.Node[i].nn; l++ ){
			
			if (g.Node[i].neigh[l] == e){
				if ( g.Edge[g.Node[i].neigh[l]].s1 == i )
					sito = g.Edge[g.Node[i].neigh[l]].s2;
				else
					sito = g.Edge[g.Node[i].neigh[l]].s1;
			
			g.Node[i].father=sito;}			
			
			
			
			if (g.Node[i].neigh[l] != e){
				if ( g.Edge[g.Node[i].neigh[l]].s1 == i )
					sito = g.Edge[g.Node[i].neigh[l]].s2;
				else
					sito = g.Edge[g.Node[i].neigh[l]].s1;
				g.Node[i].sons[nson]=sito;
				nson=nson+1;
		

				build_family( g, sito, g.Node[i].neigh[l],vecleaves,nleaves );
					}
			
				}
			
			
		}
		


	return;
}
/*********************************************/


//COMPUTE_PROBABILITIES
void compute_probabilities(Graph g, int *nin, int *listin, int *n1, int *list1, int *n2, int *list2, int *n3, int *list3){
	int i,j;
	double *dist_medie;
	double min;
       
	dist_medie=(double*)malloc(*nin * sizeof(double));
	
	//computing average distance of each leaves
	for(i=0;i<*nin;i++){
		dist_medie[i]=0.;
		for(j=0;j<*n1;j++){dist_medie[i]+=g.leavesdist[listin[i]][list1[j]];}
		
		for(j=0;j<*n2;j++){dist_medie[i]+=g.leavesdist[listin[i]][list2[j]];}

		for(j=0;j<*n3;j++){dist_medie[i]+=g.leavesdist[listin[i]][list3[j]];}

		dist_medie[i]/=(*n1+*n2+*n3);
		
	}
	
	//finding the minimum average distance
	min=dist_medie[0];
	for(j=1;j<*nin;j++){if(dist_medie[j]<min){min=dist_medie[j];}}

	//filling probabilities vector

	for(i=0;i<*nin;i++){g.Node[listin[i]].prob=pow((min/dist_medie[i]),10);}
        free(dist_medie);
	return;
}


/*********************************************/	



//COMPUTE_PROB_NODE
void compute_prob_node(Graph g, int i){

int k,j=0,num;

int *list;

list=(int*)malloc(g.nleaves*sizeof(int));
for(j=0;j<2;j++){
num=0;
ComputeLeavesFromSite(g,g.Node[i].sons[j],list,&num);
for(k=0;k<num;k++){g.Node[i].weights[j]*=(1-g.Node[list[k]].prob);}
}
g.Node[i].weights[0]+=1;
g.Node[i].weights[1]+=1;
free(list);

}
/*********************************************/	


//ASSIGN_PROB_NODES
void assign_prob_nodes(Graph g,int root){
int sito,j;
if(g.Node[root].nn==1)return;

else{	
	compute_prob_node(g,root);
 	
				for(j=0;j<2;j++){ 
                                     
					sito = g.Node[root].sons[j];
			              if(sito>=0)assign_prob_nodes( g, sito);
				        
                                       
                                                }
		
	  }
	
return;

}
/*********************************************/


//COMPUTE_LEAVES_WEIGHTS
void compute_leaves_weights(Graph g, int root, int *nin, int *listin, double *listout){

int conta,sito,direction,salva;
int j=0;
direction=salva=-1;
for(j=0;j<*nin;j++){listout[j]=1.;}

if(*nin==1){
sito=g.Node[listin[0]].father;
if(g.Node[sito].sons[0]==salva) direction=1;
if(g.Node[sito].sons[1]==salva) direction=0;
listout[0]*=g.Node[sito].weights[direction];

}

if(*nin>1){  
for(j=0;j<*nin;j++){
sito=g.Node[listin[j]].father;
conta=1;
salva=listin[j];
while(sito!=root){

if(g.Node[sito].sons[0]==salva)direction=1;
if(g.Node[sito].sons[1]==salva) direction=0;
listout[j]*=(g.Node[sito].weights[direction]/2.);
salva=sito;
sito=g.Node[sito].father;
}


if(g.Node[sito].sons[0]==salva)direction=1;
if(g.Node[sito].sons[1]==salva) direction=0;
listout[j]*=(g.Node[sito].weights[direction]/2.);

}

}

return;
}






//PRINTSUBTREE

void print_subtree(Graph g , int i)
{
	
	
	if (g.Node[i].nn == 1 ){
		printf("node: %d \n",i);
		printf("father: %d \n",g.Node[i].father);
		printf("foglia: %s prob: %f\n",g.Node[i].name,g.Node[i].prob);
		return;
	}
	
	else{	
	           printf("node: %d pb sx: %f pb dx: %f\n",i,g.Node[i].weights[0],g.Node[i].weights[1]);	
			   printf("father: %d \n",g.Node[i].father);
		       printf("sons: %d %d\n",g.Node[i].sons[0], g.Node[i].sons[1]);

					
			
				
				print_subtree( g, g.Node[i].sons[0] );
		
		print_subtree( g, g.Node[i].sons[1] );
		       
			
			
		}
		
		
	
	
	
	
	return;
}
/*********************************************/




































